# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps, loads

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import (
    Apprise,
    AppriseAsset,
    AppriseAttachment,
    NotifyFormat,
    NotifyType,
)
from apprise.plugins.telegram import NotifyTelegram

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    ##################################
    # NotifyTelegram
    ##################################
    (
        "tgram://",
        {
            "instance": None,
        },
    ),
    # Simple Message
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Simple Message (no images)
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # don't include an image by default
            "include_image": False,
        },
    ),
    # Simple Message with multiple chat names
    (
        "tgram://123456789:abcdefg_hijklmnop/id1/id2/",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Simple Message with multiple chat names
    (
        "tgram://123456789:abcdefg_hijklmnop/?to=id1,id2",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Simple Message with an invalid chat ID
    (
        "tgram://123456789:abcdefg_hijklmnop/%$/",
        {
            "instance": NotifyTelegram,
            # Notify will fail
            "response": False,
        },
    ),
    # Simple Message with multiple chat ids
    (
        "tgram://123456789:abcdefg_hijklmnop/id1/id2/23423/-30/",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Simple Message with multiple chat ids (no images)
    (
        "tgram://123456789:abcdefg_hijklmnop/id1/id2/23423/-30/",
        {
            "instance": NotifyTelegram,
            # don't include an image by default
            "include_image": False,
        },
    ),
    # Support bot keyword prefix
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Support Thread Topics
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/id1/?topic=12345",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Thread is just an alias of topic
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/id1/?thread=12345",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Threads must be numeric
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/id1/?topic=invalid",
        {
            "instance": TypeError,
        },
    ),
    # content must be 'before' or 'after'
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/id1/?content=invalid",
        {
            "instance": TypeError,
        },
    ),
    (
        "tgram://bottest@123456789:abcdefg_hijklmnop/id1:invalid/?thread=12345",
        {
            "instance": NotifyTelegram,
            # Notify will fail (bad target)
            "response": False,
        },
    ),
    # Testing image
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?image=Yes",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Testing invalid format (fall's back to html)
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=invalid",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Testing empty format (falls back to html)
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Testing valid formats
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=markdown",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=markdown&mdv=v1",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=markdown&mdv=v2",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/l2g/?format=markdown&mdv=bad",
        {
            # Defaults to v2
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=html",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?format=text",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Test Silent Settings
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?silent=yes",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?silent=no",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Test Web Page Preview Settings
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?preview=yes",
        {
            "instance": NotifyTelegram,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?preview=no",
        {
            "instance": NotifyTelegram,
        },
    ),
    # Simple Message without image
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # don't include an image by default
            "include_image": False,
        },
    ),
    # Invalid Bot Token
    (
        "tgram://alpha:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": None,
        },
    ),
    # AuthToken + bad url
    (
        "tgram://:@/",
        {
            "instance": None,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?image=Yes",
        {
            "instance": NotifyTelegram,
            # force a failure without an image specified
            "include_image": False,
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/id1/id2/",
        {
            "instance": NotifyTelegram,
            # force a failure with multiple chat_ids
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/id1/id2/",
        {
            "instance": NotifyTelegram,
            # force a failure without an image specified
            "include_image": False,
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # throw a bizarre code forcing us to fail to look it up without
            # having an image included
            "include_image": False,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    # Test with image set
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?image=Yes",
        {
            "instance": NotifyTelegram,
            # throw a bizarre code forcing us to fail to look it up without
            # having an image included
            "include_image": True,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/",
        {
            "instance": NotifyTelegram,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        "tgram://123456789:abcdefg_hijklmnop/lead2gold/?image=Yes",
        {
            "instance": NotifyTelegram,
            # Throws a series of i/o exceptions with this flag is set and
            # tests that we gracefully handle them without images set
            "include_image": True,
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_telegram_urls():
    """NotifyTelegram() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_telegram_general(mock_post):
    """NotifyTelegram() General Tests."""

    # Bot Token
    bot_token = "123456789:abcdefg_hijklmnop"
    invalid_bot_token = "abcd:123"

    # Chat ID
    chat_ids = "l2g:1234, lead2gold"

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = "{}"

    # Exception should be thrown about the fact no bot token was specified
    with pytest.raises(TypeError):
        NotifyTelegram(bot_token=None, targets=chat_ids)

    # Invalid JSON while trying to detect bot owner
    mock_post.return_value.content = "}"
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    obj.notify(title="hello", body="world")

    # Invalid JSON while trying to detect bot owner + 400 error
    mock_post.return_value.status_code = requests.codes.internal_server_error
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    obj.notify(title="hello", body="world")

    # Return status back to how they were
    mock_post.return_value.status_code = requests.codes.ok

    # Exception should be thrown about the fact an invalid bot token was
    # specifed
    with pytest.raises(TypeError):
        NotifyTelegram(bot_token=invalid_bot_token, targets=chat_ids)

    obj = NotifyTelegram(
        bot_token=bot_token, targets=chat_ids, include_image=True
    )
    assert isinstance(obj, NotifyTelegram) is True
    assert len(obj.targets) == 2

    # Test Image Sending Exceptions
    mock_post.side_effect = OSError()
    assert not obj.send_media(obj.targets[0], NotifyType.INFO)

    # Test our other objects
    mock_post.side_effect = requests.HTTPError
    assert not obj.send_media(obj.targets[0], NotifyType.INFO)

    # Restore their entries
    mock_post.side_effect = None
    mock_post.return_value.content = "{}"

    # test url call
    assert isinstance(obj.url(), str) is True

    # test privacy version of url
    assert isinstance(obj.url(privacy=True), str) is True
    assert obj.url(privacy=True).startswith("tgram://1...p/") is True

    # Test that we can load the string we generate back:
    obj = NotifyTelegram(**NotifyTelegram.parse_url(obj.url()))
    assert isinstance(obj, NotifyTelegram) is True

    # Prepare Mock to fail
    response = mock.Mock()
    response.status_code = requests.codes.internal_server_error

    # a error response
    response.content = dumps(
        {
            "description": "test",
        }
    )
    mock_post.return_value = response

    # No image asset
    nimg_obj = NotifyTelegram(bot_token=bot_token, targets=chat_ids)
    nimg_obj.asset = AppriseAsset(image_path_mask=False, image_url_mask=False)

    # Test that our default settings over-ride base settings since they are
    # not the same as the one specified in the base; this check merely
    # ensures our plugin inheritance is working properly
    assert obj.body_maxlen == NotifyTelegram.body_maxlen

    # This tests erroneous messages involving multiple chat ids
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )
    assert (
        nimg_obj.notify(
            body="body", title="title", notify_type=NotifyType.INFO
        )
        is False
    )

    # This tests erroneous messages involving a single chat id
    obj = NotifyTelegram(bot_token=bot_token, targets="l2g")
    nimg_obj = NotifyTelegram(bot_token=bot_token, targets="l2g")
    nimg_obj.asset = AppriseAsset(image_path_mask=False, image_url_mask=False)

    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )
    assert (
        nimg_obj.notify(
            body="body", title="title", notify_type=NotifyType.INFO
        )
        is False
    )

    # Bot Token Detection
    # Just to make it clear to people reading this code and trying to learn
    # what is going on.  Apprise tries to detect the bot owner if you don't
    # specify a user to message.  The idea is to just default to messaging
    # the bot owner himself (it makes it easier for people).  So we're testing
    # the creating of a Telegram Notification without providing a chat ID.
    # We're testing the error handling of this bot detection section of the
    # code
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [
                {
                    "update_id": 645421319,
                    # Entry without `message` in it
                },
                {
                    # Entry without `from` in `message`
                    "update_id": 645421320,
                    "message": {
                        "message_id": 2,
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )
    mock_post.return_value.status_code = requests.codes.ok

    obj = NotifyTelegram(bot_token=bot_token, targets="12345")
    assert len(obj.targets) == 1
    assert obj.targets[0] == (12345, None)

    # Test the escaping of characters since Telegram escapes stuff for us to
    # which we need to consider
    mock_post.reset_mock()
    body = "<p>'\"This can't\t\r\nfail&nbsp;us\"'</p>"
    assert (
        obj.notify(
            body=body, title="special characters", notify_type=NotifyType.INFO
        )
        is True
    )
    assert mock_post.call_count == 1
    payload = loads(mock_post.call_args_list[0][1]["data"])

    # Test our payload
    assert (
        payload["text"]
        == "<b>special characters</b>\r\n'\"This can't\t\r\nfail us\"'\r\n"
    )

    for content in ("before", "after"):
        # Test our content settings
        obj = NotifyTelegram(
            bot_token=bot_token, targets="12345", content=content
        )
        # Reset our mock
        mock_post.reset_mock()
        # Test sending attachments
        attach = AppriseAttachment(
            os.path.join(TEST_VAR_DIR, "apprise-test.gif")
        )
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is True
        )

        # Test large messages
        assert (
            obj.notify(
                body="a" * (obj.telegram_caption_maxlen + 1),
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is True
        )

        # An invalid attachment will cause a failure
        path = os.path.join(
            TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg"
        )
        attach = AppriseAttachment(path)
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=path,
            )
            is False
        )

        # Test large messages
        assert (
            obj.notify(
                body="a" * (obj.telegram_caption_maxlen + 1),
                title="title",
                notify_type=NotifyType.INFO,
                attach=path,
            )
            is False
        )

    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    # No user detected; this happens after our firsst notification
    assert len(obj.targets) == 0

    assert obj.notify(title="hello", body="world") is True
    assert len(obj.targets) == 1
    assert obj.targets[0] == ("532389719", None)

    # Do the test again, but without the expected (parsed response)
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [],
        }
    )

    # No user will be detected now
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    # No user detected; this happens after our firsst notification
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # Do the test again, but with ok not set to True
    mock_post.return_value.content = dumps(
        {
            "ok": False,
            "result": [
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )

    # No user will be detected now
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    # No user detected; this happens after our firsst notification
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # An edge case where no results were provided; this will probably never
    # happen, but it helps with test coverage completeness
    mock_post.return_value.content = dumps(
        {
            "ok": True,
        }
    )

    # No user will be detected now
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    # No user detected; this happens after our firsst notification
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0
    # Detect the bot with a bad response
    mock_post.return_value.content = dumps({})
    obj.detect_bot_owner()

    # Test our bot detection with a internal server error
    mock_post.return_value.status_code = requests.codes.internal_server_error

    # internal server error prevents notification from being sent
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # Test our bot detection with an unmappable html error
    mock_post.return_value.status_code = 999
    NotifyTelegram(bot_token=bot_token, targets=None)
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # Do it again but this time provide a failure message
    mock_post.return_value.content = dumps({"description": "Failure Message"})
    NotifyTelegram(bot_token=bot_token, targets=None)
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # Do it again but this time provide a failure message and perform a
    # notification without a bot detection by providing at least 1 chat id
    obj = NotifyTelegram(bot_token=bot_token, targets=["@abcd"])
    assert (
        nimg_obj.notify(
            body="body", title="title", notify_type=NotifyType.INFO
        )
        is False
    )

    # iterate over our exceptions and test them
    mock_post.side_effect = requests.HTTPError

    # No chat_ids specified
    obj = NotifyTelegram(bot_token=bot_token, targets=None)
    assert len(obj.targets) == 0
    assert obj.notify(title="hello", body="world") is False
    assert len(obj.targets) == 0

    # Test Telegram Group
    obj = Apprise.instantiate(
        "tgram://123456789:ABCdefghijkl123456789opqyz/-123456789525"
    )
    assert isinstance(obj, NotifyTelegram)
    assert len(obj.targets) == 1
    assert (-123456789525, None) in obj.targets


@mock.patch("requests.post")
def test_plugin_telegram_formatting(mock_post):
    """NotifyTelegram() formatting tests."""

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    mock_post.return_value.content = "{}"

    # Simple success response
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )
    mock_post.return_value.status_code = requests.codes.ok

    results = NotifyTelegram.parse_url("tgram://123456789:abcdefg_hijklmnop/")

    instance = NotifyTelegram(**results)
    assert isinstance(instance, NotifyTelegram)

    response = instance.send(title="title", body="body")
    assert response is True
    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/sendMessage"
    )

    # Reset our values
    mock_post.reset_mock()

    # Now test our HTML Conversion as TEXT)
    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/")
    assert len(aobj) == 1

    title = "🚨 Change detected for <i>Apprise Test Title</i>"
    body = (
        '<a href="http://localhost"><i>Apprise Body Title</i></a>'
        ' had <a href="http://127.0.0.1">a change</a>'
    )

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a TEXT mode
    assert (
        payload["text"]
        == "<b>🚨 Change detected for &lt;i&gt;Apprise Test Title&lt;/i&gt;"
        '</b>\r\n&lt;a href="http://localhost"&gt;&lt;i&gt;'
        "Apprise Body Title&lt;/i&gt;&lt;/a&gt; had &lt;"
        'a href="http://127.0.0.1"&gt;a change&lt;/a&gt;'
    )

    # Reset our values
    mock_post.reset_mock()

    # Now test our HTML Conversion as TEXT)
    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?format=html")
    assert len(aobj) == 1

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.HTML)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"]
        == "<b>🚨 Change detected for <i>Apprise Test Title</i></b>\r\n"
        '<a href="http://localhost"><i>Apprise Body Title</i></a> had '
        '<a href="http://127.0.0.1">a change</a>'
    )

    # Reset our values
    mock_post.reset_mock()

    # Now test our MARKDOWN Handling
    title = "# 🚨 Change detected for _Apprise Test Title_"
    body = (
        "_[Apprise Body Title](http://localhost)_"
        " had [a change](http://127.0.0.1)"
    )

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?format=markdown&mdv=2")
    assert len(aobj) == 1

    assert aobj.notify(
        title=title, body=body, body_format=NotifyFormat.MARKDOWN
    )

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"] == "# 🚨 Change detected for _Apprise Test Title_\r\n"
        "_[Apprise Body Title](http://localhost)_ had "
        "[a change](http://127.0.0.1)"
    )

    # Reset our values
    mock_post.reset_mock()

    # Now test our MARKDOWN Handling
    title = "# 🚨 Change detected for _Apprise Test Title_"
    body = (
        "_[Apprise Body Title](http://localhost)_"
        " had [a change](http://127.0.0.1)"
    )

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?format=markdown&mdv=1")
    assert len(aobj) == 1

    assert aobj.notify(
        title=title, body=body, body_format=NotifyFormat.MARKDOWN
    )

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot123456789:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"] == "# 🚨 Change detected for _Apprise Test Title_\r\n"
        "_[Apprise Body Title](http://localhost)_ had "
        "[a change](http://127.0.0.1)"
    )

    # Reset our values
    mock_post.reset_mock()

    # Upstream to use HTML but input specified as Markdown
    aobj = Apprise()
    aobj.add("tgram://987654321:abcdefg_hijklmnop/?format=html")
    assert len(aobj) == 1

    # Now test our MARKDOWN Handling
    title = "# 🚨 Another Change detected for _Apprise Test Title_"
    body = (
        "_[Apprise Body Title](http://localhost)_"
        " had [a change](http://127.0.0.2)"
    )

    # HTML forced by the command line, but MARKDOWN specified as
    # upstream mode
    assert aobj.notify(
        title=title, body=body, body_format=NotifyFormat.MARKDOWN
    )

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"] == "<b>\r\n<b>🚨 Another Change detected for "
        "<i>Apprise Test Title</i></b>\r\n</b>\r\n<i>"
        '<a href="http://localhost">Apprise Body Title</a>'
        '</i> had <a href="http://127.0.0.2">a change</a>\r\n'
    )

    # Now we'll test an edge case where a title was defined, but after
    # processing it, it was determiend there really wasn't anything there
    # at all at the end of the day.

    # Reset our values
    mock_post.reset_mock()

    # Upstream to use HTML but input specified as Markdown v1
    aobj = Apprise()
    aobj.add("tgram://987654321:abcdefg_hijklmnop/?format=markdown&mdv=1")
    assert len(aobj) == 1

    # Now test our MARKDOWN Handling (no title defined... not really anyway)
    title = "# "
    body = (
        "_[Apprise Body Title](http://localhost)_"
        " had [a change](http://127.0.0.2)"
    )

    # MARKDOWN forced by the command line, but TEXT specified as
    # upstream mode
    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a MARKDOWN mode
    assert payload["text"] == body

    # Reset our values
    mock_post.reset_mock()

    # Upstream to use HTML but input specified as Markdown v2
    aobj = Apprise()
    aobj.add("tgram://987654321:abcdefg_hijklmnop/?format=markdown&mdv=2")
    assert len(aobj) == 1

    # MARKDOWN forced by the command line, but TEXT specified as
    # upstream mode
    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a MARKDOWN mode
    assert (
        payload["text"]
        == "\\_\\[Apprise Body Title\\]\\(http://localhost\\)\\_ had \\"
        "[a change\\]\\(http://127\\.0\\.0\\.2\\)"
    )

    # Reset our values
    mock_post.reset_mock()

    # Upstream to use HTML but input specified as Markdown v1
    aobj = Apprise()
    aobj.add("tgram://987654321:abcdefg_hijklmnop/?format=markdown&mdv=1")
    assert len(aobj) == 1

    # Set an actual title this time
    title = "# A Great Title"
    body = (
        "_[Apprise Body Title](http://localhost)_"
        " had [a change](http://127.0.0.2)"
    )

    # TEXT forced by the command line, but MARKDOWN specified as
    # upstream mode
    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a MARKDOWN mode
    assert (
        payload["text"] == "# A Great Title\r\n"
        "_[Apprise Body Title](http://localhost)_ had "
        "[a change](http://127.0.0.2)"
    )

    # Reset our values
    mock_post.reset_mock()

    # Upstream to use HTML but input specified as Markdown v2
    aobj = Apprise()
    aobj.add("tgram://987654321:abcdefg_hijklmnop/?format=markdown&mdv=2")
    assert len(aobj) == 1

    # TEXT forced by the command line, but MARKDOWN specified as
    # upstream mode
    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 2

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/getUpdates"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a MARKDOWN mode
    assert (
        payload["text"] == "\\# A Great Title\r\n"
        "\\_\\[Apprise Body Title\\]\\(http://localhost\\)\\_ had "
        "\\[a change\\]\\(http://127\\.0\\.0\\.2\\)"
    )

    # Reset our values
    mock_post.reset_mock()

    # If input is markdown and output is v2, it is expected the user knows
    # what he is doing... no esaping takes place
    assert aobj.notify(
        title=title, body=body, body_format=NotifyFormat.MARKDOWN
    )

    # Test our calls
    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])

    # No escaping in this circumstance
    assert (
        payload["text"] == "# A Great Title\r\n"
        "_[Apprise Body Title](http://localhost)_ had "
        "[a change](http://127.0.0.2)"
    )

    # Reset our values
    mock_post.reset_mock()

    # No body format specified at all... user definitely must know what
    # they are doing... still no escaping in this circumstance
    assert aobj.notify(title=title, body=body)

    # Test our calls
    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot987654321:abcdefg_hijklmnop/sendMessage"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])

    # No escaping in this circumstance
    assert (
        payload["text"] == "# A Great Title\r\n"
        "_[Apprise Body Title](http://localhost)_ had "
        "[a change](http://127.0.0.2)"
    )

    # Reset our values
    mock_post.reset_mock()

    #
    # Now test that <br/> is correctly escaped
    #
    title = "Test Message Title"
    body = "Test Message Body <br/> ok</br>"

    aobj = Apprise()
    aobj.add("tgram://1234:aaaaaaaaa/-1123456245134")
    assert len(aobj) == 1

    assert aobj.notify(
        title=title, body=body, body_format=NotifyFormat.MARKDOWN
    )

    # Test our calls
    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot1234:aaaaaaaaa/sendMessage"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"]
        == "<b>Test Message Title\r\n</b>\r\nTest Message Body\r\nok\r\n"
    )

    # Reset our values
    mock_post.reset_mock()

    #
    # Now test that <br/> is correctly escaped as it would have been via the
    # CLI mode where the body_format is TEXT
    #

    aobj = Apprise()
    aobj.add("tgram://1234:aaaaaaaaa/-1123456245134")
    assert len(aobj) == 1

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # Test our calls
    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot1234:aaaaaaaaa/sendMessage"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"] == "<b>Test Message Title</b>\r\n"
        "Test Message Body &lt;br/&gt; ok&lt;/br&gt;"
    )

    # Reset our values
    mock_post.reset_mock()

    #
    # Now test that <br/> is correctly escaped if fed as HTML
    #

    aobj = Apprise()
    aobj.add("tgram://1234:aaaaaaaaa/-1123456245134")
    assert len(aobj) == 1

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.HTML)

    # Test our calls
    assert mock_post.call_count == 1

    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.telegram.org/bot1234:aaaaaaaaa/sendMessage"
    )

    payload = loads(mock_post.call_args_list[0][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"]
        == "<b>Test Message Title</b>\r\nTest Message Body\r\nok\r\n"
    )


@mock.patch("requests.post")
def test_plugin_telegram_html_formatting(mock_post):
    """NotifyTelegram() HTML Formatting."""
    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Simple success response
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/")

    assert len(aobj) == 1

    assert isinstance(aobj[0], NotifyTelegram)

    # Test our HTML Conversion
    title = "<title>&apos;information&apos</title>"
    body = (
        "<em>&quot;This is in Italic&quot</em><br/>"
        "<h5>&emsp;&emspHeadings&nbsp;are dropped and"
        "&nbspconverted to bold</h5>"
    )

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.HTML)

    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2

    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Test that everything is escaped properly in a HTML mode
    assert (
        payload["text"]
        == "<b>\r\n<b>'information'</b>\r\n</b>\r\n<i>\"This is in Italic\""
        "</i>\r\n<b>      Headings are dropped and converted to bold</b>\r\n"
    )

    mock_post.reset_mock()

    assert aobj.notify(title=title, body=body, body_format=NotifyFormat.TEXT)

    # owner has already been looked up, so only one call is made
    assert mock_post.call_count == 1

    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert (
        payload["text"]
        == "<b>&lt;title&gt;&amp;apos;information&amp;apos&lt;/title&gt;</b>"
        "\r\n&lt;em&gt;&amp;quot;This is in Italic&amp;quot&lt;/em&gt;&lt;"
        "br/&gt;&lt;h5&gt;&amp;emsp;&amp;emspHeadings&amp;nbsp;are "
        "dropped and&amp;nbspconverted to bold&lt;/h5&gt;"
    )

    # Lest test more complex HTML examples now
    mock_post.reset_mock()

    test_file_01 = os.path.join(TEST_VAR_DIR, "01_test_example.html")
    with open(test_file_01) as html_file:
        assert aobj.notify(
            body=html_file.read(), body_format=NotifyFormat.HTML
        )

    # owner has already been looked up, so only one call is made
    assert mock_post.call_count == 1

    payload = loads(mock_post.call_args_list[0][1]["data"])
    assert (
        payload["text"]
        == "\r\n<b>Bootstrap 101 Template</b>\r\n<b>My Title</b>\r\n"
        "<b>Heading 1</b>\r\n-Bullet 1\r\n-Bullet 2\r\n-Bullet 3\r\n"
        "-Bullet 1\r\n-Bullet 2\r\n-Bullet 3\r\n<b>Heading 2</b>\r\n"
        "A div entry\r\nA div entry\r\n"
        "<pre><code class=\"language-python\">print('hello')</code></pre>\r\n"
        "<b>Heading 3</b>\r\n<b>Heading 4</b>\r\n<b>Heading 5</b>\r\n"
        "<b>Heading 6</b>\r\nA set of text\r\n"
        "Another line after the set of text\r\nMore text\r\nlabel"
    )


@mock.patch("requests.post")
def test_plugin_telegram_threads(mock_post):
    """NotifyTelegram() Threads/Topics."""
    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Simple success response
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?thread=1234")

    assert len(aobj) == 1

    assert isinstance(aobj[0], NotifyTelegram)

    body = "my threaded message"

    assert aobj.notify(body=body)

    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2

    payload = loads(mock_post.call_args_list[1][1]["data"])

    assert "message_thread_id" in payload
    assert payload["message_thread_id"] == 1234

    mock_post.reset_mock()

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?topic=1234")

    assert len(aobj) == 1

    assert isinstance(aobj[0], NotifyTelegram)

    body = "my message"

    assert aobj.notify(body=body)

    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2

    payload = loads(mock_post.call_args_list[1][1]["data"])

    assert "message_thread_id" in payload
    assert payload["message_thread_id"] == 1234

    mock_post.reset_mock()

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/9876:1234/9876:1111")

    assert len(aobj) == 1

    assert isinstance(aobj[0], NotifyTelegram)

    body = "my message"

    assert aobj.notify(body=body)

    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2

    payload = loads(mock_post.call_args_list[0][1]["data"])

    assert "message_thread_id" in payload
    assert payload["message_thread_id"] == 1111

    payload = loads(mock_post.call_args_list[1][1]["data"])

    assert "message_thread_id" in payload
    assert payload["message_thread_id"] == 1234

    mock_post.reset_mock()


@mock.patch("requests.post")
def test_plugin_telegram_markdown_v2(mock_post):
    """NotifyTelegram() MarkdownV2."""
    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Simple success response
    mock_post.return_value.content = dumps(
        {
            "ok": True,
            "result": [
                {
                    "update_id": 645421321,
                    "message": {
                        "message_id": 2,
                        "from": {
                            "id": 532389719,
                            "is_bot": False,
                            "first_name": "Chris",
                            "language_code": "en-US",
                        },
                        "chat": {
                            "id": 532389719,
                            "first_name": "Chris",
                            "type": "private",
                        },
                        "date": 1519694394,
                        "text": "/start",
                        "entities": [
                            {
                                "offset": 0,
                                "length": 6,
                                "type": "bot_command",
                            }
                        ],
                    },
                },
            ],
        }
    )

    aobj = Apprise()
    aobj.add("tgram://123456789:abcdefg_hijklmnop/?mdv=2&format=markdown")
    assert len(aobj) == 1
    assert isinstance(aobj[0], NotifyTelegram)

    body = "# my message\r\n## more content\r\n\\# already escaped hashtag"

    # Test with body format set to markdown
    assert aobj.notify(body=body, body_format=NotifyFormat.TEXT)

    # 1 call to look up bot owner, and second for notification
    assert mock_post.call_count == 2
    payload = loads(mock_post.call_args_list[1][1]["data"])

    # Our content is escapped properly
    assert (
        payload["text"] == "\\# my message\r\n"
        "\\#\\# more content\r\n\\# already escaped hashtag"
    )

    mock_post.reset_mock()

    # We'll iterate over all of the bad unsupported characters
    mdv2_unsupported = (
        "_",
        "*",
        "[",
        "]",
        "(",
        ")",
        "~",
        "`",
        ">",
        "#",
        "+",
        "=",
        "|",
        "{",
        "}",
        ".",
        "!",
        "-",
    )

    for c in mdv2_unsupported:
        body = f"bad character: {c}, and already escapped \\{c}"

        # Test with body format set to markdown
        assert aobj.notify(body=body, body_format=NotifyFormat.TEXT)
        assert mock_post.call_count == 1
        payload = loads(mock_post.call_args_list[0][1]["data"])

        # Our content is escapped properly
        assert (
            payload["text"]
            == f"bad character: \\{c}, and already escapped \\{c}"
        )

        mock_post.reset_mock()


@mock.patch("requests.post")
def test_plugin_telegram_attach_memory(mock_post):
    """Regression: AttachMemory must be sendable without OSError."""
    from apprise.attachment.memory import AttachMemory

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = dumps({"ok": True, "result": True})
    mock_post.return_value = response

    obj = NotifyTelegram(
        bot_token="123456789:abcdefg_hijklmnop", targets="12345"
    )

    mem = AttachMemory(
        content=b"<html><body><h1>Test</h1></body></html>",
        name="test.html",
        mimetype="text/html",
    )

    assert obj.notify(body="Test", attach=mem) is True
    assert mock_post.call_count >= 1



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Great Resources:
# - Dev/Legacy API:
#    https://firebase.google.com/docs/cloud-messaging/http-server-ref
# - Legacy API (v1) -> OAuth
# - https://firebase.google.com/docs/cloud-messaging/migrate-v1

import json
import os
import sys
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise
from apprise.plugins.fcm import NotifyFCM

try:
    from cryptography.exceptions import UnsupportedAlgorithm

    from apprise.plugins.fcm.color import FCMColorManager
    from apprise.plugins.fcm.common import FCM_MODES
    from apprise.plugins.fcm.oauth import GoogleOAuth
    from apprise.plugins.fcm.priority import FCM_PRIORITIES, FCMPriorityManager

except ImportError:
    # No problem; there is no cryptography support
    pass


# Disable logging for a cleaner testing output
import logging

logging.disable(logging.CRITICAL)

# Test files for KeyFile Directory
PRIVATE_KEYFILE_DIR = os.path.join(os.path.dirname(__file__), "var", "fcm")
FCM_KEYFILE = os.path.join(PRIVATE_KEYFILE_DIR, "service_account.json")


# Our Testing URLs
apprise_url_tests = (
    (
        "fcm://",
        {
            # We failed to identify any valid authentication
            "instance": TypeError,
        },
    ),
    (
        "fcm://:@/",
        {
            # We failed to identify any valid authentication
            "instance": TypeError,
        },
    ),
    (
        "fcm://project@%20%20/",
        {
            # invalid apikey
            "instance": TypeError,
        },
    ),
    (
        "fcm://apikey/",
        {
            # no project id specified so we operate in legacy mode
            "instance": NotifyFCM,
            # but there are no targets specified so we return False
            "notify_response": False,
        },
    ),
    (
        "fcm://apikey/device",
        {
            # Valid device
            "instance": NotifyFCM,
            "privacy_url": "fcm://a...y/device",
        },
    ),
    (
        "fcm://apikey/#topic",
        {
            # Valid topic
            "instance": NotifyFCM,
            "privacy_url": "fcm://a...y/%23topic",
        },
    ),
    (
        "fcm://apikey/device?mode=invalid",
        {
            # Valid device, invalid mode
            "instance": TypeError,
        },
    ),
    (
        "fcm://apikey/#topic1/device/%20/",
        {
            # Valid topic, valid device, and invalid entry
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://apikey?to=#topic1,device",
        {
            # Test to=
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://?apikey=abc123&to=device",
        {
            # Test apikey= to=
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://?apikey=abc123&to=device&image=yes",
        {
            # Test image boolean
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://?apikey=abc123&to=device&color=no",
        {
            # Disable colors
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://?apikey=abc123&to=device&color=aabbcc",
        {
            # custom colors
            "instance": NotifyFCM,
        },
    ),
    (
        (
            "fcm://?apikey=abc123&to=device"
            "&image_url=http://example.com/interesting.jpg"
        ),
        {
            # Test image_url
            "instance": NotifyFCM
        },
    ),
    (
        (
            "fcm://?apikey=abc123&to=device"
            "&image_url=http://example.com/interesting.jpg&image=no"
        ),
        {
            # Test image_url but set to no
            "instance": NotifyFCM
        },
    ),
    (
        "fcm://?apikey=abc123&to=device&+key=value&+key2=value2",
        {
            # Test apikey= to= and data arguments
            "instance": NotifyFCM,
        },
    ),
    (
        "fcm://%20?to=device&keyfile=/invalid/path",
        {
            # invalid Project ID
            "instance": TypeError,
        },
    ),
    (
        "fcm://project_id?to=device&keyfile=/invalid/path",
        {
            # Test to= and auto-detection of OAuth mode
            "instance": NotifyFCM,
            # we'll fail to send our notification as a result
            "response": False,
        },
    ),
    (
        "fcm://?to=device&project=project_id&keyfile=/invalid/path",
        {
            # Test project= & to= and auto detection of OAuth mode
            "instance": NotifyFCM,
            # we'll fail to send our notification as a result
            "response": False,
        },
    ),
    (
        "fcm://project_id?to=device&mode=oauth2",
        {
            # no keyfile was specified
            "instance": TypeError,
        },
    ),
    (
        "fcm://project_id?to=device&mode=oauth2&keyfile=/invalid/path",
        {
            # Same test as above except we explicitly set our oauth2 mode
            # Test to= and auto-detection of OAuth mode
            "instance": NotifyFCM,
            # we'll fail to send our notification as a result
            "response": False,
        },
    ),
    (
        "fcm://apikey/#topic1/device/?mode=legacy",
        {
            "instance": NotifyFCM,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "fcm://apikey/#topic1/device/?mode=legacy",
        {
            "instance": NotifyFCM,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        "fcm://project/#topic1/device/?mode=oauth2&keyfile=file://{}".format(
            os.path.join(
                os.path.dirname(__file__), "var", "fcm", "service_account.json"
            )
        ),
        {
            "instance": NotifyFCM,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "fcm://projectid/#topic1/device/?mode=oauth2&keyfile=file://{}".format(
            os.path.join(
                os.path.dirname(__file__), "var", "fcm", "service_account.json"
            )
        ),
        {
            "instance": NotifyFCM,
            # Throws a series of connection and transfer exceptions when
            # this flag is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


@pytest.fixture
def mock_post(mocker):
    """Prepare a good OAuth mock response."""

    mock_thing = mocker.patch("requests.post")

    response = mock.Mock()
    response.content = json.dumps(
        {
            "access_token": "ya29.c.abcd",
            "expires_in": 3599,
            "token_type": "Bearer",
        }
    )
    response.status_code = requests.codes.ok
    mock_thing.return_value = response

    return mock_thing


@pytest.fixture
def mock_post_legacy(mocker):
    """Prepare a good legacy mock response."""

    mock_thing = mocker.patch("requests.post")

    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""
    mock_thing.return_value = response

    return mock_thing


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_urls():
    """NotifyFCM() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_legacy_default(mock_post_legacy):
    """NotifyFCM() Legacy/APIKey default checks."""

    # A valid Legacy URL
    obj = Apprise.instantiate(
        "fcm://abc123/device/"
        "?+key=value&+key2=value2"
        "&image_url=https://example.com/interesting.png"
    )

    # Send our notification
    assert obj.notify("test") is True

    # Test our call count
    assert mock_post_legacy.call_count == 1
    assert (
        mock_post_legacy.call_args_list[0][0][0]
        == "https://fcm.googleapis.com/fcm/send"
    )

    payload = mock_post_legacy.mock_calls[0][2]
    data = json.loads(payload["data"])
    assert "data" in data
    assert isinstance(data, dict)
    assert "key" in data["data"]
    assert data["data"]["key"] == "value"
    assert "key2" in data["data"]
    assert data["data"]["key2"] == "value2"

    assert "notification" in data
    assert isinstance(data["notification"], dict)
    assert "notification" in data["notification"]
    assert isinstance(data["notification"]["notification"], dict)
    assert "image" in data["notification"]["notification"]
    assert (
        data["notification"]["notification"]["image"]
        == "https://example.com/interesting.png"
    )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_legacy_priorities(mock_post_legacy):
    """NotifyFCM() Legacy/APIKey priorities checks."""

    obj = Apprise.instantiate("fcm://abc123/device/?priority=low")
    assert mock_post_legacy.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post_legacy.call_count == 1
    assert (
        mock_post_legacy.call_args_list[0][0][0]
        == "https://fcm.googleapis.com/fcm/send"
    )

    payload = mock_post_legacy.mock_calls[0][2]
    data = json.loads(payload["data"])
    assert "data" not in data
    assert "notification" in data
    assert isinstance(data["notification"], dict)
    assert "notification" in data["notification"]
    assert isinstance(data["notification"]["notification"], dict)
    assert "image" not in data["notification"]["notification"]
    assert "priority" in data

    # legacy can only switch between high/low
    assert data["priority"] == "normal"


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_legacy_no_colors(mock_post_legacy):
    """NotifyFCM() Legacy/APIKey `color=no` checks."""

    obj = Apprise.instantiate("fcm://abc123/device/?color=no")
    assert mock_post_legacy.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post_legacy.call_count == 1
    assert (
        mock_post_legacy.call_args_list[0][0][0]
        == "https://fcm.googleapis.com/fcm/send"
    )

    payload = mock_post_legacy.mock_calls[0][2]
    data = json.loads(payload["data"])
    assert "data" not in data
    assert "notification" in data
    assert isinstance(data["notification"], dict)
    assert "notification" in data["notification"]
    assert isinstance(data["notification"]["notification"], dict)
    assert "image" not in data["notification"]["notification"]
    assert "color" not in data["notification"]["notification"]


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_legacy_colors(mock_post_legacy):
    """NotifyFCM() Legacy/APIKey colors checks."""

    obj = Apprise.instantiate("fcm://abc123/device/?color=AA001b")
    assert mock_post_legacy.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post_legacy.call_count == 1
    assert (
        mock_post_legacy.call_args_list[0][0][0]
        == "https://fcm.googleapis.com/fcm/send"
    )

    payload = mock_post_legacy.mock_calls[0][2]
    data = json.loads(payload["data"])
    assert "data" not in data
    assert "notification" in data
    assert isinstance(data["notification"], dict)
    assert "notification" in data["notification"]
    assert isinstance(data["notification"]["notification"], dict)
    assert "image" not in data["notification"]["notification"]
    assert "color" in data["notification"]["notification"]
    assert data["notification"]["notification"]["color"] == "#aa001b"


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_default(mock_post):
    """
    NotifyFCM() general OAuth checks - success.
    Test using a valid Project ID and key file.
    """

    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/#topic/?keyfile={FCM_KEYFILE}"
    )

    # send our notification
    assert obj.notify("test") is True

    # Test our call count
    assert mock_post.call_count == 3
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_invalid_project_id(mock_post):
    """NotifyFCM() OAuth checks, with invalid project id."""

    # Test having a valid keyfile, but not a valid project id match.
    obj = Apprise.instantiate(
        f"fcm://invalid_project_id/device/?keyfile={FCM_KEYFILE}"
    )

    # we'll fail as a result
    assert obj.notify("test") is False

    # Test our call count
    assert mock_post.call_count == 0


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_keyfile_error(mock_post):
    """NotifyFCM() OAuth checks, while unable to read key file."""

    # Now we test using a valid Project ID but we can't open our file
    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/?keyfile={FCM_KEYFILE}"
    )

    with mock.patch("builtins.open", side_effect=OSError):
        # we'll fail as a result
        assert obj.notify("test") is False

    # Test our call count
    assert mock_post.call_count == 0


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_data_parameters(mock_post):
    """NotifyFCM() OAuth checks, success.

    Test using a valid Project ID and data parameters.
    """

    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/#topic/?keyfile={FCM_KEYFILE}"
        "&+key=value&+key2=value2"
        "&image_url=https://example.com/interesting.png"
    )
    assert mock_post.call_count == 0

    # send our notification
    assert obj.notify("test") is True

    # Test our call count
    assert mock_post.call_count == 3
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )
    payload = mock_post.mock_calls[1][2]
    data = json.loads(payload["data"])
    assert "message" in data
    assert isinstance(data["message"], dict)
    assert "data" in data["message"]
    assert isinstance(data["message"]["data"], dict)
    assert "key" in data["message"]["data"]
    assert data["message"]["data"]["key"] == "value"
    assert "key2" in data["message"]["data"]
    assert data["message"]["data"]["key2"] == "value2"

    assert "notification" in data["message"]
    assert isinstance(data["message"]["notification"], dict)
    assert "image" in data["message"]["notification"]
    assert (
        data["message"]["notification"]["image"]
        == "https://example.com/interesting.png"
    )

    assert (
        mock_post.call_args_list[2][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )

    payload = mock_post.mock_calls[2][2]
    data = json.loads(payload["data"])
    assert "message" in data
    assert isinstance(data["message"], dict)
    assert "data" in data["message"]
    assert isinstance(data["message"]["data"], dict)
    assert "key" in data["message"]["data"]
    assert data["message"]["data"]["key"] == "value"
    assert "key2" in data["message"]["data"]
    assert data["message"]["data"]["key2"] == "value2"

    assert "notification" in data["message"]
    assert isinstance(data["message"]["notification"], dict)
    assert "image" in data["message"]["notification"]
    assert (
        data["message"]["notification"]["image"]
        == "https://example.com/interesting.png"
    )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_priorities(mock_post):
    """Verify priorities work as intended."""

    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/?keyfile={FCM_KEYFILE}&priority=high"
    )
    assert mock_post.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )
    payload = mock_post.mock_calls[1][2]
    data = json.loads(payload["data"])
    assert "message" in data
    assert isinstance(data["message"], dict)
    assert "data" not in data["message"]
    assert "notification" in data["message"]
    assert isinstance(data["message"]["notification"], dict)
    assert "image" not in data["message"]["notification"]
    assert data["message"]["apns"]["headers"]["apns-priority"] == "10"
    assert data["message"]["webpush"]["headers"]["Urgency"] == "high"
    assert data["message"]["android"]["priority"] == "HIGH"


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_no_colors(mock_post):
    """Verify `color=no` work as intended."""

    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/?keyfile={FCM_KEYFILE}&color=no"
    )
    assert mock_post.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )
    payload = mock_post.mock_calls[1][2]
    data = json.loads(payload["data"])
    assert "message" in data
    assert isinstance(data["message"], dict)
    assert "data" not in data["message"]
    assert "notification" in data["message"]
    assert isinstance(data["message"]["notification"], dict)
    assert "color" not in data["message"]["notification"]


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_oauth_colors(mock_post):
    """Verify colors work as intended."""

    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/?keyfile={FCM_KEYFILE}&color=#12AAbb"
    )
    assert mock_post.call_count == 0

    # Send our notification
    assert obj.notify(title="title", body="body") is True

    # Test our call count
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )

    assert (
        mock_post.call_args_list[1][0][0]
        == "https://fcm.googleapis.com/v1/projects/mock-project-id/messages:send"
    )
    payload = mock_post.mock_calls[1][2]
    data = json.loads(payload["data"])
    assert "message" in data
    assert isinstance(data["message"], dict)
    assert "data" not in data["message"]
    assert "notification" in data["message"]
    assert isinstance(data["message"]["notification"], dict)
    assert "color" in data["message"]["android"]["notification"]
    assert data["message"]["android"]["notification"]["color"] == "#12aabb"


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_parse_default(mock_post):
    """NotifyFCM() KeyFile Tests."""

    oauth = GoogleOAuth()
    # We can not get an Access Token without content loaded
    assert oauth.access_token is None

    # Load our content
    assert oauth.load(FCM_KEYFILE) is True
    assert oauth.access_token is not None

    # Test our call count
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )

    mock_post.reset_mock()

    # a second call uses cache since our token hasn't expired yet
    assert oauth.access_token is not None
    assert mock_post.call_count == 0


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_parse_no_expiry(mock_post):
    """Test case without `expires_in` entry."""

    mock_post.return_value.content = json.dumps(
        {
            "access_token": "ya29.c.abcd",
            "token_type": "Bearer",
        }
    )

    oauth = GoogleOAuth()
    assert oauth.load(FCM_KEYFILE) is True
    assert oauth.access_token is not None

    # Test our call count
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_parse_user_agent(mock_post):
    """Test case with `user-agent` override."""

    oauth = GoogleOAuth(user_agent="test-agent-override")
    assert oauth.load(FCM_KEYFILE) is True
    assert oauth.access_token is not None
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://accounts.google.com/o/oauth2/token"
    )


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_parse_keyfile_failures(mock_post: mock.Mock):
    """Test some errors that can get thrown when trying to handle the
    `service_account.json` file."""

    # Now we test a case where we can't access the file we've been pointed to:
    oauth = GoogleOAuth()
    with mock.patch("builtins.open", side_effect=OSError):
        # We will fail to retrieve our Access Token
        assert oauth.load(FCM_KEYFILE) is False
        assert oauth.access_token is None

    oauth = GoogleOAuth()
    with mock.patch("json.loads", side_effect=([],)):
        # We will fail to retrieve our Access Token since we did not parse
        # a dictionary
        assert oauth.load(FCM_KEYFILE) is False
        assert oauth.access_token is None

    # Case where we can't load the PEM key:
    oauth = GoogleOAuth()
    with mock.patch(
        "cryptography.hazmat.primitives.serialization.load_pem_private_key",
        side_effect=ValueError(""),
    ):
        assert oauth.load(FCM_KEYFILE) is False
        assert oauth.access_token is None

    # Case where we can't load the PEM key:
    oauth = GoogleOAuth()
    with mock.patch(
        "cryptography.hazmat.primitives.serialization.load_pem_private_key",
        side_effect=TypeError(""),
    ):
        assert oauth.load(FCM_KEYFILE) is False
        assert oauth.access_token is None

    # Case where we can't load the PEM key:
    oauth = GoogleOAuth()
    with mock.patch(
        "cryptography.hazmat.primitives.serialization.load_pem_private_key",
        side_effect=UnsupportedAlgorithm(""),
    ):
        # Note: This test should be te
        assert oauth.load(FCM_KEYFILE) is False
        assert oauth.access_token is None

    # Verify that not a single call to the web escaped the test harness.
    assert mock_post.mock_calls == []


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_parse_token_failures(mock_post):
    """Test some web errors that can occur when speaking upstream with Google
    to get our token generated."""

    mock_post.return_value.status_code = requests.codes.internal_server_error
    mock_post.return_value.content = b""

    oauth = GoogleOAuth()
    assert oauth.load(FCM_KEYFILE) is True

    # We'll fail due to an bad web response
    assert oauth.access_token is None

    # Return our status code to how it was
    mock_post.return_value.status_code = requests.codes.ok

    # No access token
    bad_response_1 = mock.Mock()
    bad_response_1.content = json.dumps(
        {
            "expires_in": 3599,
            "token_type": "Bearer",
        }
    )

    # Invalid JSON
    bad_response_2 = mock.Mock()
    bad_response_2.content = "{"

    mock_post.return_value = None
    # Throw an exception on the first call to requests.post()
    for side_effect in (
        requests.RequestException(),
        [bad_response_1],
        [bad_response_2],
    ):
        mock_post.side_effect = side_effect

        # Test all of our bad side effects
        oauth = GoogleOAuth()
        assert oauth.load(FCM_KEYFILE) is True

        # We'll fail due to an bad web response
        assert oauth.access_token is None


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_bad_keyfile_parse():
    """NotifyFCM() KeyFile Bad Service Account Type Tests."""

    path = os.path.join(PRIVATE_KEYFILE_DIR, "service_account-bad-type.json")
    oauth = GoogleOAuth()
    assert oauth.load(path) is False


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_keyfile_missing_entries_parse(tmpdir):
    """NotifyFCM() KeyFile Missing Entries Test."""

    # Prepare a base keyfile reference to use
    path = os.path.join(PRIVATE_KEYFILE_DIR, "service_account.json")
    with open(path, encoding="utf-8") as fp:
        content = json.loads(fp.read())

    path = tmpdir.join("fcm_keyfile.json")

    # Test that we fail to load if the following keys are missing:
    for entry in (
        "client_email",
        "private_key_id",
        "private_key",
        "type",
        "project_id",
    ):
        # Ensure the key actually exists in our file
        assert entry in content

        # Create a copy of our content
        content_copy = content.copy()

        # Remove our entry we expect to validate against
        del content_copy[entry]
        assert entry not in content_copy

        path.write(json.dumps(content_copy))

        oauth = GoogleOAuth()
        assert oauth.load(str(path)) is False

    # Now write ourselves a bad JSON file
    path.write("{")
    oauth = GoogleOAuth()
    assert oauth.load(str(path)) is False


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_priority_manager():
    """NotifyFCM() FCMPriorityManager() Testing."""

    for mode in FCM_MODES:
        for priority in FCM_PRIORITIES:
            instance = FCMPriorityManager(mode, priority)
            assert isinstance(instance.payload(), dict)
            # Verify it's not empty
            assert bool(instance)
            assert instance.payload()
            assert str(instance) == priority

    # We do not have to set a priority
    instance = FCMPriorityManager(mode)
    assert isinstance(instance.payload(), dict)

    # Dictionary is empty though
    assert not bool(instance)
    assert not instance.payload()
    assert str(instance) == ""

    with pytest.raises(TypeError):
        instance = FCMPriorityManager(mode, "invalid")

    with pytest.raises(TypeError):
        instance = FCMPriorityManager("invald", "high")

    # mode validation is done at the higher NotifyFCM() level so
    # it is not tested here (not required)


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
def test_plugin_fcm_color_manager():
    """NotifyFCM() FCMColorManager() Testing."""

    # No colors
    instance = FCMColorManager("no")
    assert bool(instance) is False
    assert instance.get() is None
    # We'll return that we are not defined
    assert str(instance) == "no"

    # Asset colors
    instance = FCMColorManager("yes")
    assert isinstance(instance.get(), str)
    # Output: #rrggbb
    assert len(instance.get()) == 7
    # Starts with has symbol
    assert instance.get()[0] == "#"
    # We'll return that we are defined but using default configuration
    assert str(instance) == "yes"

    # We will be `true` because we can acquire a color based on what was
    # passed in
    assert bool(instance)

    # Custom color
    instance = FCMColorManager("#A2B3A4")
    assert isinstance(instance.get(), str)
    assert instance.get() == "#a2b3a4"
    assert bool(instance)
    # str() response does not include hashtag
    assert str(instance) == "a2b3a4"

    # Custom color (no hashtag)
    instance = FCMColorManager("A2B3A4")
    assert isinstance(instance.get(), str)
    # Hashtag is always part of output
    assert instance.get() == "#a2b3a4"
    assert bool(instance)
    # str() response does not include hashtag
    assert str(instance) == "a2b3a4"

    # Custom color (no hashtag) but only using 3 letter rgb values
    instance = FCMColorManager("AC4")
    assert isinstance(instance.get(), str)
    # Hashtag is always part of output
    assert instance.get() == "#aacc44"
    assert bool(instance)
    # str() response does not include hashtag
    assert str(instance) == "aacc44"


@pytest.mark.skipif(
    "cryptography" in sys.modules,
    reason="Requires that cryptography NOT be installed",
)
def test_plugin_fcm_cryptography_import_error():
    """NotifyFCM Cryptography loading failure."""

    # Prepare a base keyfile reference to use
    path = os.path.join(PRIVATE_KEYFILE_DIR, "service_account.json")

    # Attempt to instantiate our object
    obj = Apprise.instantiate(
        f"fcm://mock-project-id/device/#topic/?keyfile={path!s}"
    )

    # It's not possible because our cryptography depedancy is missing
    assert obj is None


@pytest.mark.skipif(
    "cryptography" not in sys.modules, reason="Requires cryptography"
)
@mock.patch("requests.post")
def test_plugin_fcm_edge_cases(mock_post):
    """NotifyFCM() Edge Cases."""

    # Prepare a good response
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b""
    mock_post.return_value = response

    # this tests an edge case where verify if the data_kwargs is a dictionary
    # or not.  Below, we don't even define it, so it will be None (causing
    # the check to go).  We'll still correctly instantiate a plugin:
    obj = NotifyFCM("project", "api:123", targets="device")
    assert obj is not None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise.plugins.sfr import NotifySFR

logging.disable(logging.CRITICAL)

SFR_GOOD_RESPONSE = json.dumps(
    {
        "success": True,
        "reponse": 8888888,
    }
)

SFR_BAD_RESPONSE = json.dumps(
    {
        "success": False,
        "errorCode": "THIS_IS_AN_ERROR",
        "errorDetail": "Appel api en erreur",
        "fatal": True,
        "invalidParams": True,
    }
)

# Our Testing URLs
apprise_url_tests = (
    (
        "sfr://",
        {
            # No host specified
            "instance": TypeError,
        },
    ),
    (
        "sfr://:@/",
        {
            # Invalid host
            "instance": TypeError,
        },
    ),
    (
        "sfr://:service_password",
        {
            # No user specified
            "instance": TypeError,
        },
    ),
    (
        "sfr://testing:serv@ice_password",
        {
            # Invalid Password
            "instance": TypeError,
        },
    ),
    (
        "sfr://testing:service_password@/5555555555",
        {
            # No spaceId provided
            "instance": TypeError,
        },
    ),
    (
        "sfr://testing:service_password@12345/",
        {
            # No target provided
            "instance": TypeError,
        },
    ),
    (
        f"sfr://:service_password@12345/{3 * 13}",
        {
            # No host but everything else provided
            "instance": TypeError,
        },
    ),
    (
        "sfr://:service_password@space_id/targets?media=TEST",
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:",
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:@",
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:@{}".format("0" * 3),
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:@{}/".format("0" * 3),
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:@{}/targets".format("0" * 3),
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:@{}/targets?media=TEST".format("0" * 3),
        {
            "instance": TypeError,
        },
    ),
    (
        "sfr://service_id:service_password@{}/{}?from=MyApp&timeout=30".format(
            "0" * 3, "0" * 10
        ),
        {
            # a valid group
            "instance": NotifySFR,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "sfr://service_id:****@0...0/0000000000?"
                "from=MyApp&timeout=30&voice=claire08s&"
                "lang=fr_FR&media=SMSUnicode"
            ),
            # Our response expected server response
            "requests_response_text": SFR_GOOD_RESPONSE,
        },
    ),
    (
        "sfr://service_id:service_password@{}/{}?voice=laura8k&lang=en_US".format(
            "0" * 3, "0" * 10
        ),
        {
            # a valid group
            "instance": NotifySFR,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "sfr://service_id:****@0...0/0000000000?"
                "from=&timeout=2880&voice=laura8k&"
                "lang=en_US&media=SMSUnicode"
            ),
            # Our response expected server response
            "requests_response_text": SFR_GOOD_RESPONSE,
        },
    ),
    (
        "sfr://service_id:service_password@{}/{}?media=SMS".format(
            "0" * 3, "0" * 10
        ),
        {
            # a valid group
            "instance": NotifySFR,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "sfr://service_id:****@0...0/0000000000?"
                "from=&timeout=2880&voice=claire08s&"
                "lang=fr_FR&media=SMS"
            ),
            # Our response expected server response
            "requests_response_text": SFR_GOOD_RESPONSE,
        },
    ),
    (
        "sfr://service_id:service_password@{}/{}".format("0" * 3, "0" * 10),
        {
            # Test case where we get a bad response
            "instance": NotifySFR,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "sfr://service_id:****@0...0/0000000000?"
                "from=&timeout=2880&voice=claire08s&"
                "lang=fr_FR&media=SMSUnicode"
            ),
            # Our failed notification expected server response
            "requests_response_text": SFR_BAD_RESPONSE,
            "requests_response_code": requests.codes.ok,
            # as a result, we expect a failed notification
            "response": False,
        },
    ),
)


def test_plugin_sfr_urls():
    """NotifySFR() Apprise URLs."""
    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_sfr_notification_ok(mock_post):
    """NotifySFR() Notifications Ok response."""
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = SFR_GOOD_RESPONSE
    mock_post.return_value = response

    # Test our URL parsing
    results = NotifySFR.parse_url(
        "sfr://srv:pwd@{}/{}?media=SMSLong".format("1" * 8, "0" * 10)
    )

    assert isinstance(results, dict)
    assert results["user"] == "srv"
    assert results["password"] == "pwd"
    assert results["space_id"] == "11111111"
    assert results["targets"] == ["0000000000"]
    assert results["media"] == "SMSLong"
    assert results["timeout"] == ""
    assert results["voice"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "fr_FR"
    assert instance.lang == "fr_FR"
    assert instance.sender == ""
    assert isinstance(instance.targets, list)
    assert isinstance(instance.timeout, int)
    assert isinstance(instance.voice, str)
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    assert response is True
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_sfr_notification_multiple_targets_ok(mock_post):
    """NotifySFR() Notifications ko response."""
    # Reset our object
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = SFR_GOOD_RESPONSE
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:other_fjv&8password@{}/?to={},{}&from=MyCustomUser".format(
            "4" * 6, "1" * 8, "6" * 10, "8" * 10
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "444444"
    assert results["password"] == "other_fjv&8password"
    assert results["space_id"] == "11111111"
    assert results["targets"] == ["6666666666", "8888888888"]
    assert results["media"] == ""
    assert results["timeout"] == ""
    assert results["voice"] == ""
    assert results["lang"] == ""
    assert results["sender"] == "MyCustomUser"

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 2
    assert instance.lang == "fr_FR"
    assert instance.sender == "MyCustomUser"
    assert instance.media == "SMSUnicode"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 2880
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    assert response is True
    assert mock_post.call_count == 2


@mock.patch("requests.post")
def test_plugin_sfr_notification_ko(mock_post):
    """NotifySFR() Notifications ko response."""
    # Reset our object
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = SFR_BAD_RESPONSE
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:other_fjv&8password@{}/{}?timeout=30&media=SMS".format(
            "4" * 6, "1" * 8, "2" * 10
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "444444"
    assert results["password"] == "other_fjv&8password"
    assert results["space_id"] == "11111111"
    assert results["media"] == "SMS"
    assert results["targets"] == ["2222222222"]
    assert results["timeout"] == "30"
    assert results["voice"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "fr_FR"
    assert instance.sender == ""
    assert instance.media == "SMS"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 30
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    assert response is False
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_sfr_notification_multiple_targets_all_ko(mock_post):
    """NotifySFR() Notifications ko response."""
    # Reset our object
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = SFR_BAD_RESPONSE
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:other_fjv&8password@{}/?to={},{}&voice=laura8k".format(
            "4" * 6, "1" * 8, "6" * 4, "8" * 4
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "444444"
    assert results["password"] == "other_fjv&8password"
    assert results["space_id"] == "11111111"
    assert results["targets"] == ["6666", "8888"]
    assert results["voice"] == "laura8k"
    assert results["media"] == ""
    assert results["timeout"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    # No valid phone number provided
    with pytest.raises(TypeError):
        NotifySFR(**results)


@mock.patch("requests.post")
def test_plugin_sfr_notification_multiple_targets_one_ko(mock_post):
    """NotifySFR() Notifications ko response."""
    # Reset our object
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = SFR_BAD_RESPONSE
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:&pass@{}/?to={},{}&media=SMSUnicodeLong&lang=en_US".format(
            "4" * 6, "1" * 8, "6" * 10, "8" * 4
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "444444"
    assert results["password"] == "&pass"
    assert results["space_id"] == "11111111"
    assert results["targets"] == ["6666666666", "8888"]
    assert results["voice"] == ""
    assert results["media"] == "SMSUnicodeLong"
    assert results["timeout"] == ""
    assert results["lang"] == "en_US"
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "en_US"
    assert instance.sender == ""
    assert instance.media == "SMSUnicodeLong"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 2880
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    # One phone number failed to be parsed, therefore notify fails
    response = instance.send(body="test")
    assert response is False
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_sfr_notification_exceptions(mock_post):
    """NotifySFR() Notifications exceptions."""
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.internal_server_error
    response.content = SFR_GOOD_RESPONSE
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:str0*fn_ppw0rd@{}/{}".format(
            "404ghwo89144", "9993384", "0959290404"
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "404ghwo89144"
    assert results["password"] == "str0*fn_ppw0rd"
    assert results["space_id"] == "9993384"
    assert results["targets"] == ["0959290404"]
    assert results["media"] == ""
    assert results["timeout"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "fr_FR"
    assert instance.sender == ""
    assert instance.media == "SMSUnicode"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 2880
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    # Must return False
    assert response is False
    assert mock_post.call_count == 1

    # Test invalid content returned by requests
    mock_post.reset_mock()
    response = mock.Mock()
    response.status_code = requests.codes.ok
    response.content = b"Invalid JSON Content"
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:str0*fn_ppw0rd@{}/{}".format(
            "404ghwo89144", "9993384", "0959290404"
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "404ghwo89144"
    assert results["password"] == "str0*fn_ppw0rd"
    assert results["space_id"] == "9993384"
    assert results["targets"] == ["0959290404"]
    assert results["media"] == ""
    assert results["timeout"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "fr_FR"
    assert instance.sender == ""
    assert instance.media == "SMSUnicode"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 2880
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    # Must return False
    assert response is False
    assert mock_post.call_count == 1


@mock.patch(
    "requests.post",
    side_effect=requests.RequestException("Connection error"),
)
def test_plugin_sfr_notification_exceptions_requests(mock_post):
    """NotifySFR() Notifications requests exceptions."""
    # Test requests socket error return
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.internal_server_error
    response.content = b"Invalid content"
    mock_post.return_value = response

    # Test "real" parameters
    results = NotifySFR.parse_url(
        "sfr://{}:str0*fn_ppw0rd@{}/{}".format(
            "404ghwo89144", "9993384", "0959290404"
        )
    )

    assert isinstance(results, dict)
    assert results["user"] == "404ghwo89144"
    assert results["password"] == "str0*fn_ppw0rd"
    assert results["space_id"] == "9993384"
    assert results["targets"] == ["0959290404"]
    assert results["media"] == ""
    assert results["timeout"] == ""
    assert results["lang"] == ""
    assert results["sender"] == ""

    instance = NotifySFR(**results)
    assert isinstance(instance, NotifySFR)
    assert len(instance) == 1
    assert instance.lang == "fr_FR"
    assert instance.sender == ""
    assert instance.media == "SMSUnicode"
    assert isinstance(instance.targets, list)
    assert instance.timeout == 2880
    assert instance.voice == "claire08s"
    assert isinstance(instance.space_id, str)

    response = instance.send(body="test")
    # Must return False do to requests error
    assert response is False
    assert mock_post.call_count == 1


@mock.patch("requests.post")
def test_plugin_sfr_failure(mock_post):
    """NotifySFR() Failure Cases."""
    mock_post.reset_mock()
    # Prepare Mock
    # Create a mock response object
    response = mock.Mock()
    response.status_code = requests.codes.no_content
    mock_post.return_value = response

    # Invalid service_id
    with pytest.raises(TypeError):
        NotifySFR(
            user=None,
            password="service_password",
            space_id=int("8" * 10),
            targets=int("8" * 10),
        )

    # Invalid service_password
    with pytest.raises(TypeError):
        NotifySFR(
            user="service_id",
            password=None,
            space_id=int("8" * 10),
            targets=int("8" * 10),
        )

    # Invalid space_id
    with pytest.raises(TypeError):
        NotifySFR(
            user="service_id",
            password="service_password",
            space_id=None,
            targets=int("8" * 10),
        )

    # Invalid targets
    with pytest.raises(TypeError):
        NotifySFR(
            user="service_id",
            password="service_password",
            space_id=int("8" * 10),
            targets=None,
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import re
import urllib

import pytest

from apprise import AppriseAttachment, exception
from apprise.attachment.base import AttachBase
from apprise.attachment.memory import AttachMemory
from apprise.common import ContentLocation

logging.disable(logging.CRITICAL)


def test_attach_memory_parse_url():
    """
    API: AttachMemory().parse_url()

    """

    # Bad Entry
    assert AttachMemory.parse_url(object) is None

    # Our filename is detected automatically
    assert AttachMemory.parse_url("memory://")

    # pass our content in as a string
    mem = AttachMemory(content="string")
    # it loads a string type by default
    assert mem.mimetype == "text/plain"
    # Our filename is automatically generated (with .txt)
    assert re.match(r"^[a-z0-9-]+\.txt$", mem.name, re.I)

    # open our file
    with mem as fp:
        assert fp.getbuffer().nbytes == len(mem)

    # pass our content in as a string
    mem = AttachMemory(
        content="<html/>", name="test.html", mimetype="text/html"
    )
    # it loads a string type by default
    assert mem.mimetype == "text/html"
    assert mem.name == "test.html"

    # Stub function
    assert mem.download()

    with pytest.raises(TypeError):
        # garbage in, garbage out
        AttachMemory(content=3)

    # pointer to our data
    pointer = mem.open()
    assert pointer.read() == b"<html/>"

    # pass our content in as a string
    mem = AttachMemory(content=b"binary-data", name="raw.dat")
    # it loads a string type by default
    assert mem.mimetype == "application/octet-stream"
    assert mem.name == "raw.dat"

    # pass our content in as a string
    mem = AttachMemory(content=b"binary-data")
    # it loads a string type by default
    assert mem.mimetype == "application/octet-stream"
    # Our filename is automatically generated (with .dat)
    assert re.match(r"^[a-z0-9-]+\.dat$", mem.name, re.I)


def test_attach_memory():
    """
    API: AttachMemory()

    """
    # A url we can test with
    fname = "testfile"
    url = f"memory:///ignored/path/{fname}"

    # Simple gif test
    response = AppriseAttachment.instantiate(url)
    assert isinstance(response, AttachMemory)

    # There is no path yet as we haven't written anything to our memory object
    # yet
    assert response.path is None
    assert bool(response) is False

    with response as memobj:
        memobj.write(b"content")

    # Memory object defaults
    assert response.name == fname
    assert response.path == response.name
    assert response.mimetype == "application/octet-stream"
    assert bool(response) is True

    #
    fname_in_url = urllib.parse.quote(response.name)
    assert response.url().startswith(f"memory://{fname_in_url}")

    # Mime is always part of url
    assert re.search(r"[?&]mime=", response.url()) is not None

    # Test case where location is simply set to INACCESSIBLE
    # Below is a bad example, but it proves the section of code properly works.
    # Ideally a server admin may wish to just disable all File based
    # attachments entirely. In this case, they simply just need to change the
    # global singleton at the start of their program like:
    #
    # import apprise
    # apprise.attachment.AttachMemory.location = \
    #       apprise.ContentLocation.INACCESSIBLE
    #
    response = AppriseAttachment.instantiate(url)
    assert isinstance(response, AttachMemory)
    with response as memobj:
        memobj.write(b"content")

    response.location = ContentLocation.INACCESSIBLE
    assert response.path is None
    # Downloads just don't work period
    assert response.download() is False

    # File handling (even if image is set to maxium allowable)
    response = AppriseAttachment.instantiate(url)
    assert isinstance(response, AttachMemory)
    with response as memobj:
        memobj.write(b"content")

    # Memory handling when size is to large
    response = AppriseAttachment.instantiate(url)
    assert isinstance(response, AttachMemory)
    with response as memobj:
        memobj.write(b"content")

    # Test case where we exceed our defined max_file_size in memory
    prev_value = AttachBase.max_file_size
    AttachBase.max_file_size = len(response) - 1
    # We can't work in this case
    assert response.path is None
    assert response.download() is False

    # Restore our file_size
    AttachBase.max_file_size = prev_value

    response = AppriseAttachment.instantiate(
        "memory://apprise-file.gif?mime=image/gif"
    )
    assert isinstance(response, AttachMemory)
    with response as memobj:
        memobj.write(b"content")

    assert response.name == "apprise-file.gif"
    assert response.path == response.name
    assert response.mimetype == "image/gif"
    # No mime-type and/or filename over-ride was specified, so therefore it
    # won't show up in the generated URL
    assert re.search(r"[?&]mime=", response.url()) is not None
    assert "image/gif" in response.url()

    # Force a mime-type and new name
    response = AppriseAttachment.instantiate(
        "memory://{}?mime={}&name={}".format(
            "ignored.gif", "image/jpeg", "test.jpeg"
        )
    )
    assert isinstance(response, AttachMemory)
    with response as memobj:
        memobj.write(b"content")

    assert response.name == "test.jpeg"
    assert response.path == response.name
    assert response.mimetype == "image/jpeg"
    # We will match on mime type now  (%2F = /)
    assert re.search(r"[?&]mime=image/jpeg", response.url(), re.I)
    assert response.url().startswith("memory://test.jpeg")

    # Test hosted configuration and that we can't add a valid memory file
    aa = AppriseAttachment(location=ContentLocation.HOSTED)
    assert aa.add(response) is False

    # now test our base64 output
    assert isinstance(response.base64(), str)
    # No encoding if we choose
    assert isinstance(response.base64(encoding=None), bytes)

    response.invalidate()
    with pytest.raises(exception.AppriseFileNotFound):
        response.base64()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import locale

# Disable logging for a cleaner testing output
import logging
import os

logging.disable(logging.CRITICAL)


@contextlib.contextmanager
def environ(*remove, **update):
    """Temporarily updates the ``os.environ`` dictionary in-place.

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.

    :param remove: Environment variable(s) to remove.
    :param update: Dictionary of environment variables and values to
                   add/update.
    """
    env_orig = os.environ.copy()
    loc_orig = locale.getlocale()
    try:
        os.environ.update(update)
        for k in remove:
            os.environ.pop(k, None)
        yield
    finally:
        os.environ.clear()
        os.environ.update(env_orig)
        with contextlib.suppress(locale.Error):
            locale.setlocale(locale.LC_ALL, loc_orig)
